'use strict';

exports.__esModule = true;
exports.impliedRemove = exports.removeAndShape = exports.remover = exports.keepAndShape = exports.keeper = undefined;

var _reject = require('ramda/src/reject');

var _reject2 = _interopRequireDefault(_reject);

var _omit = require('ramda/src/omit');

var _omit2 = _interopRequireDefault(_omit);

var _is = require('ramda/src/is');

var _is2 = _interopRequireDefault(_is);

var _filter = require('ramda/src/filter');

var _filter2 = _interopRequireDefault(_filter);

var _pick = require('ramda/src/pick');

var _pick2 = _interopRequireDefault(_pick);

var _keys = require('ramda/src/keys');

var _keys2 = _interopRequireDefault(_keys);

var _pipe = require('ramda/src/pipe');

var _pipe2 = _interopRequireDefault(_pipe);

var _curry = require('ramda/src/curry');

var _curry2 = _interopRequireDefault(_curry);

var _evolveSpec = require('./evolveSpec');

var _evolveSpec2 = _interopRequireDefault(_evolveSpec);

var _util = require('./util');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * Keeps only the props you name in your spec.
 * It is expected (best practice, for conveying intent) that you give a value of `true` for props that you want to keep
 * but _not_ transform (a zipped object - where key and value are both strings of the same name - makes sense too).
 *
 * @func
 * @sig {k: v} -> {k: v} -> {k: v}
 * @param {Object} spec An object whose values are `true` or match the key name
 * to signal a value to be preserved (as-is) on the output
 * @param {Object} input An object which will be passed through the pruning defined by the spec
 * @returns {Object} The input object with the unnamed props shaved off the output
 */
var keeper = exports.keeper = (0, _curry2.default)(function (spec, input) {
  return (0, _pipe2.default)(_util.objectify, (0, _pipe2.default)(_util.makePruningSpec, _keys2.default, _pick2.default)(spec))(input);
});

/**
 * Keeps only the props you name in your spec. If you provided a function, it is applied at the prop-level.
 * It is expected (best practice, for conveying intent) that you give a value of `true` for props that you want to keep
 * but _not_ transform (a zipped object - where key and value are both strings of the same name - makes sense too).
 * In this mode, if you provide any other type of value it won't be used in the spec.
 * If the prop _doesn't_ exist, you won't see it on the output
 * (use shapeSuperStrictly if you want to see `undefined` for those cases instead).
 *
 * @func
 * @sig {k: v} -> {k: v} -> {k: v}
 * @param {Object} spec An object whose values are either transform functions or (preferrably)
 * a value of `true` to signal a value to be preserved (as-is) on the output
 * @param {Object} input An object which will be passed through the pruning/re-shaping defined by the spec
 * @returns {Object} The input object with the spec transforms applied to it and
 * unnamed props shaved off the output
 */
var keepAndShape = exports.keepAndShape = (0, _curry2.default)(function (spec, input) {
  return (0, _pipe2.default)(keeper(spec), (0, _evolveSpec2.default)((0, _filter2.default)((0, _is2.default)(Function), spec)))(input);
});

/**
 * Removes the props you name in your spec.
 * It is expected (best practice, for conveying intent) that you give a value of `true` for props that you want to remove
 * but _not_ transform (a zipped object - where key and value are both strings of the same name - makes sense too).
 *
 * @func
 * @sig {k: v} -> {k: v} -> {k: v}
 * @param {Object} spec An object whose values are `true` or match the key name
 * to signal a value to be removed from the output
 * @param {Object} input An object which will be passed through the pruning defined by the spec
 * @returns {Object} The input object with the named props shaved off the output
 */
var remover = exports.remover = (0, _curry2.default)(function (spec, input) {
  return (0, _pipe2.default)(_util.objectify, (0, _pipe2.default)(_util.makePruningSpec, (0, _reject2.default)((0, _is2.default)(Function)), _keys2.default, _omit2.default)(spec))(input);
});

/**
 * Removes all the props you name in your spec. If you provided a transform function however,
 * it is applied at the prop-level and that prop is (obviously) NOT removed but rather transformed accordingly.
 * It is expected (best practice, for conveying intent) that you give a value of `true` for props you want to remove
 * (a zipped object - where key and value are both strings of the same name - makes sense too).
 * In this mode, if you provide any other type of value it won't be used in the spec.
 *
 * @func
 * @sig {k: v} -> {k: v} -> {k: v}
 * @param {Object} spec An object whose values are either transform functions or (preferrably)
 * a value of `true` to signal a value to be remove from the output
 * @param {Object} input An object which will be passed through the pruning/re-shaping defined by the spec
 * @returns {Object} The input object with the spec transforms applied to it and
 * name props shaved off the output
 */
var removeAndShape = exports.removeAndShape = (0, _curry2.default)(function (spec, input) {
  return (0, _pipe2.default)(remover(spec), (0, _evolveSpec2.default)((0, _filter2.default)((0, _is2.default)(Function), spec)))(input);
});

/**
 * Removes the any props from the input not named in the spec. Doesn't matter
 * what the value types are in the spec: if the key name doesn't exist in the
 * spec, it won't exist in the output.
 *
 * @func
 * @sig {k: v} -> {k: v} -> {k: v}
 * @param {Object} spec An object
 * @param {Object} input An object which will have any props not named in the spect removed
 * @returns {Object} The input object with the props removed which were not in the spec too
 */
var impliedRemove = exports.impliedRemove = (0, _curry2.default)(function (spec, input) {
  return (0, _pick2.default)((0, _keys2.default)(spec), input);
});