import _default24 from 'ramda/src/omit';
import _default23 from 'ramda/src/tryCatch';
import _default22 from 'ramda/src/mapObjIndexed';
import _default21 from 'ramda/src/identity';
import _default20 from 'ramda/src/T';
import _default19 from 'ramda/src/equals';
import _default18 from 'ramda/src/cond';
import _default17 from 'ramda/src/prop';
import _default16 from 'ramda/src/assoc';
import _default15 from 'ramda/src/converge';
import _default14 from 'ramda/src/merge';
import _default13 from 'ramda/src/curry';
import _default12 from 'ramda/src/test';
import _default11 from 'ramda/src/reject';
import _default10 from 'ramda/src/when';
import _default9 from 'ramda/src/filter';
import _default8 from 'ramda/src/fromPairs';
import _default7 from 'ramda/src/toPairs';
import _default6 from 'ramda/src/pipe';
import _default5 from 'ramda/src/either';
import _default4 from 'ramda/src/is';
import _default3 from 'ramda/src/always';
import _default2 from 'ramda/src/unless';
import _default from 'ramda/src/pathEq';


/**
 * This is what you really mean when you're checking to see if something is an Object.
 * Honestly who cares that an Array, Date and RegExp are technically objects too?
 * You were looking for the thing with the curly braces {} right?
 *
 * @func
 * @sig * -> Boolean
 * @param {*} val A value of any type
 * @returns {Boolean} Whether or not the value is an Object
 */
export var isPlainObj = _default(['constructor', 'name'], 'Object');

/**
 * Turns any non-Object values into empty objects 
 *
 * @func
 * @sig * -> {k: v}
 * @param {*} val A value of any type
 * @returns {Object} If the original value was an Object, it is returned as-is,
 * otherwise an empty object is returned
 */
export var objectify = _default2(isPlainObj, _default3({}));

/**
 * If the provided value is NOT a function, transform it into a function that
 * always returns that value 
 *
 * @func
 * @sig * -> (* -> *)
 * @param {*} val A value of any type
 * @returns {Function} If the original value was a function, that function is returned,
 * otherwise a function that always returns the original value is returned
 * function that always returns that value OR 
 */
export var alwaysFunction = _default2(_default4(Function), _default3);

/**
 * If the provided value is neither an Object nor a Function, transform it into
 * a function that always returns that provided value.
 *
 * @func
 * @sig * -> (* -> *)|{k: v}
 * @param {*} val A value of any type
 * @returns {Function|Object} A function that always returns the provided value
 * OR the original value (if it was already a Function or an Object)
 */
export var makeFunctionUnlessObject = _default2(_default5(_default4(Function), isPlainObj), _default3);

/**
 * Makes sure a given spec is acceptable for one of the pruning modes (ie, "remove" or "keep"). 
 * Specs in this mode are treated more strictly, meaning the prop must be given
 * a value of "true" or the key and value can be identical.
 * Functions are acceptable values in the spec of course.
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} spec A spec to be coerced into an acceptable pruning spec
 * @returns {Object} A spec that is acceptable to be used in pruning mode.
 */
export var makePruningSpec = _default6(objectify, _default7, _default9(function (_ref) {
  var key = _ref[0],
      val = _ref[1];
  return key === val || val === true || _default4(Function, val);
}), _default8);

/**
 * Filters a spec (object) to only the props that are transform functions. If
 * the input is not an object, this is just an identity function. 
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} spec An object whose values (may) be transform functions
 * @returns {Object} The input object with only the props that are functions retained
 */
export var onlySpecTransforms = _default10(isPlainObj, _default6(_default7, _default9(function (_ref2) {
  var key = _ref2[0],
      val = _ref2[1];
  return key === 'shapeyDebug' || _default4(Function, val);
}), _default8));

/**
 * Removes the reserved "shapey*" prefixed props from a spec.
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} spec A shapey spec to be pruned of magic props
 * @returns {Object} A shapey spec cleaned of magic props
 */
export var removeMagicProps = _default6(_default7, _default11(function (_ref3) {
  var key = _ref3[0];
  return _default12(/^shapey/i, key);
}), _default8);

/**
 * Takes an non-transform props (non-functions) in a given spec and merges them onto the input object.
 * 
 * @func
 * @sig {k: v} -> {k: v} -> {k: v}
 * @param {Object} spec An object whose values are either transform functions or pass-through values to be added to the return object
 * @param {Object} input An object which will be passed through the re-shaping transform functions defined by the spec
 * @returns {Object} The input object with only the pass-through props applied to it
 */
export var applyNonTransformProps = _default13(function (spec, value) {
  return [spec, value].every(isPlainObj) ? _default6(removeMagicProps, _default11(_default4(Function)), _default14(value))(spec) : value;
});

/**
 * Logs to the console any failed transform functions, along with the field name
 * and the value that was fed into the transform function. Plus, the exception
 * itself is also logged.
 * 
 * @func
 * @sig String -> * -> Error -> *
 * @param {String} fieldName The field name for the failed transform function
 * @param {Object} value The value that was fed into the transform function
 * @param {Object} exception The exception thrown by the failed transform
 * @returns {undefined}
 */
var defaultErrorHandler = _default13(function (fieldName, value, exception) {
  return (
    // eslint-disable-next-line no-console
    console.error('\n Transform failed on field: "' + fieldName + '"\n value:', value, '\n', exception)
  );
});

/**
 * A wrapper around an (optional) error handler that the may be passed in as the
 * "shapeyDebug" value. It is curried, but will feed their handler a more
 * standard signature of: (err, field, value)
 *
 * @func
 * @sig ((Error, String, *) -> *) -> String -> * -> Error -> *
 * @param {Function} errHandler The custom error handler provided by the consumer
 * @param {String} fieldName The field name for the failed transform function
 * @param {Object} value The value that was fed into the transform function
 * @param {Object} exception The exception thrown by the failed transform
 * @returns {*} Could be anything, or nothing (it's up to the consumer)
 */
var wrapperForTheirErrorHandler = _default13(function (errHandler, fieldName, value, exception) {
  return errHandler(exception, fieldName, value);
});

/**
 * Wraps every function on a given spec in a try/catch that catches any exception.
 *
 * What it _does_ with the exception is up to the consumer:
 *   - ignores it (if "shapeyDebug" isn't set)
 *   - logs it along with the corresponding field name,
 *     using `console.error` (if "shapeyDebug" is set to `true`)
 *   - uses a custom handler supplied by the consumer as the value for "shapeyDebug" 
 *
 * Note that a custom error handler will be passed the following params (in order):
 *   - The exception
 *   - The field name
 *   - The value that was fed into the transform function
 *
 * @func
 * @sig {k: (a -> b)} -> {k: (a -> b)}
 * @param {Object} spec An object whose values may be transform functions that
 * need to be wrapped in try/catch
 * @returns {Object} The same spec object but whose functions are
 * safely wrapped in in try/catch handlers
 */
export var safeSpecTransforms = _default6(objectify, _default15(_default16('shapeyDebug'), [_default6(_default17('shapeyDebug'), _default18([[_default19(true), _default3(defaultErrorHandler)], [_default4(Function), wrapperForTheirErrorHandler], [_default12(/^skip$/i), function () {
  return function (_, originalValue) {
    return _default3(originalValue);
  };
}], [_default20, function () {
  return function () {
    return _default3(undefined);
  };
}]])), _default21]), _default22(function (transform, fieldName, obj) {
  return _default10(_default4(Function), _default13(function (fn, val) {
    return _default23(fn, obj.shapeyDebug(fieldName, val))(val);
  }))(transform);
}), _default24(['shapeyDebug']));